use crate::{error::ApiError, Handler};
use near_client::{core::hash::CryptoHash, near_units::parse_gas, prelude::*};
use std::collections::HashSet;
use uuid::Uuid;
use wasm_bindgen::prelude::*;

pub async fn init_meeting(
    Handler {
        contract_id,
        rpc_url,
        ..
    }: &Handler,
    signer: &Signer,
    participants: HashSet<AccountId>,
) -> Result<(Uuid, CryptoHash), JsValue> {
    let client = NearClient::new(rpc_url.clone()).map_err(ApiError::from)?;
    let meet_id = Uuid::new_v4();

    let transaction_id = client
        .function_call(signer, contract_id, "init_meeting")
        .args(serde_json::json!({
            "meet_id": meet_id,
            "participants": participants
        }))
        .gas(parse_gas!("300 T") as u64)
        .commit(Finality::None)
        .await
        .map_err(ApiError::from)?
        .id();

    Ok((meet_id, transaction_id))
}

pub async fn add_participant(
    Handler {
        contract_id,
        rpc_url,
        ..
    }: &Handler,
    signer: &Signer,
    meet_id: Uuid,
    participant: AccountId,
) -> Result<CryptoHash, ApiError> {
    let client = NearClient::new(rpc_url.clone())?;

    let transaction_id = client
        .function_call(signer, contract_id, "add_participant")
        .args(serde_json::json!({
            "meet_id": meet_id,
            "participant": participant
        }))
        .gas(parse_gas!("20 T") as u64)
        .commit(Finality::None)
        .await?
        .id();

    Ok(transaction_id)
}

pub async fn view_moderator_account(
    Handler {
        contract_id,
        rpc_url,
        ..
    }: &Handler,
    meet_id: Uuid,
) -> Result<Option<AccountId>, ApiError> {
    let client = NearClient::new(rpc_url.clone())?;

    let moderator = client
        .view::<Option<AccountId>>(
            contract_id,
            Finality::None,
            "view_moderator_account",
            Some(serde_json::json!({ "meet_id": meet_id })),
        )
        .await?
        .data();

    Ok(moderator)
}

pub async fn view_meet_participants(
    Handler {
        contract_id,
        rpc_url,
        ..
    }: &Handler,
    meet_id: Uuid,
) -> Result<Option<HashSet<AccountId>>, ApiError> {
    let client = NearClient::new(rpc_url.clone())?;

    let participants = client
        .view::<Option<HashSet<AccountId>>>(
            contract_id,
            Finality::None,
            "view_meet_participants",
            Some(serde_json::json!({ "meet_id": meet_id })),
        )
        .await?
        .data();

    Ok(participants)
}

pub async fn is_meet_participant(
    Handler {
        contract_id,
        rpc_url,
        ..
    }: &Handler,
    meet_id: Uuid,
    participant: AccountId,
) -> Result<bool, ApiError> {
    let client = NearClient::new(rpc_url.clone())?;

    let is_session_participant = *client
        .view::<bool>(
            contract_id,
            Finality::None,
            "is_meet_participant",
            Some(serde_json::json!({
                "meet_id": meet_id,
                "participant": participant
            })),
        )
        .await?;

    Ok(is_session_participant)
}

pub async fn view_server_key(
    Handler {
        contract_id,
        rpc_url,
        ..
    }: &Handler,
) -> Result<Ed25519PublicKey, ApiError> {
    let client = NearClient::new(rpc_url.clone())?;

    let public_key = *client
        .view::<Ed25519PublicKey>(contract_id, Finality::None, "view_server_key", None)
        .await?;

    Ok(public_key)
}
