use near_client::{core::account::id::ParseAccountError, crypto::Error as CryptoErr, Error};
use reqwest::Error as ExchangeError;
use serde::{Deserialize, Serialize};
use serde_json::Error as SerializationError;
use wasm_bindgen::prelude::*;

#[derive(Clone, Debug, Deserialize, Serialize)]
#[serde(tag = "error", content = "body")]
#[serde(rename_all = "snake_case")]
pub enum ApiError {
    KeyExchange(String),
    DeserializeResponse(String),
    SerializeData(String),
    NearClient(String),
    InvalidAccountId(String),
    InvalidSessionUuid(String),
    CryptoError(String),
    CreateSignatureHeader,
    VerifySignatureHeader,
    NoSignatureHeader,
    CallTimeout(String),
    AesKeyCreation(String),
    MessageEncryption(String),
    MessageDecryption(String),
    Other(String),
}

impl From<Error> for ApiError {
    fn from(err: Error) -> Self {
        Self::NearClient(err.to_string())
    }
}

impl From<ParseAccountError> for ApiError {
    fn from(err: ParseAccountError) -> Self {
        Self::InvalidAccountId(err.to_string())
    }
}

impl From<CryptoErr> for ApiError {
    fn from(err: CryptoErr) -> Self {
        Self::CryptoError(err.to_string())
    }
}

impl From<ExchangeError> for ApiError {
    fn from(err: ExchangeError) -> Self {
        Self::KeyExchange(err.to_string())
    }
}

impl From<SerializationError> for ApiError {
    fn from(err: SerializationError) -> Self {
        Self::DeserializeResponse(err.to_string())
    }
}

impl From<uuid::Error> for ApiError {
    fn from(err: uuid::Error) -> Self {
        Self::InvalidSessionUuid(err.to_string())
    }
}

impl From<ApiError> for JsValue {
    fn from(api_err: ApiError) -> Self {
        match serde_wasm_bindgen::to_value(&api_err) {
            Ok(value) => value,
            Err(err) => err.into(),
        }
    }
}
