use common_api::{
    api::{ApiResponse, Data, ExchangeMessage, ParticipantInfo, PublicKeys},
    headers::{SignatureHeader, SIGNATURE_HEADER_NAME},
};

use reqwest::{
    header::{HeaderMap, HeaderName, HeaderValue, CONTENT_TYPE},
    Client, ClientBuilder, Response,
};

use crate::{contract::view_server_key, error::ApiError, Handler};
use base64::prelude::*;
use near_client::prelude::*;
use std::{collections::HashSet, time::Duration};
use uuid::Uuid;
use wasm_bindgen::JsValue;

type Result<T> = std::result::Result<T, JsValue>;

fn client() -> Result<Client> {
    let mut headers = HeaderMap::new();
    headers.insert(CONTENT_TYPE, HeaderValue::from_static("application/json"));

    ClientBuilder::new()
        .default_headers(headers)
        .build()
        .map_err(|err| ApiError::Other(err.to_string()).into())
}

pub async fn public_keys(
    handler: &Handler,
    signer: &Signer,
    meet_id: Uuid,
    participants: HashSet<AccountId>,
) -> Result<ApiResponse<Vec<ParticipantInfo>>> {
    let client = client()?;

    let public_keys = PublicKeys {
        timeout: Duration::from_secs(10),
        participants,
    };

    let signature_header = signature_header(&public_keys, signer)?;

    // TODO: Handle Server errors
    let response = client
        .post(handler.add_path("v1/public-keys"))
        .query(&[("meet_id", meet_id)])
        .header(
            HeaderName::from_static(SIGNATURE_HEADER_NAME),
            HeaderValue::from_str(&signature_header)
                .map_err(|_| ApiError::CreateSignatureHeader)?,
        )
        .json(&public_keys)
        .send()
        .await
        .and_then(Response::error_for_status)
        .map_err(ApiError::from)?;

    let infos = serde_json::from_slice::<ApiResponse<Vec<ParticipantInfo>>>(
        &verify_response(handler, response).await?,
    )
    .map_err(ApiError::from)?;

    Ok(infos)
}

pub async fn receive(
    handler: &Handler,
    signer: &Signer,
    meet_id: Uuid,
) -> Result<ApiResponse<Data>> {
    let client = client()?;

    let duration = Duration::from_secs(10);
    let signature_header = signature_header(&duration, signer)?;

    // TODO: Handle Server errors
    let response = client
        .post(handler.add_path("v1/receive"))
        .query(&[("meet_id", meet_id)])
        .header(
            HeaderName::from_static(SIGNATURE_HEADER_NAME),
            HeaderValue::from_str(&signature_header)
                .map_err(|_| ApiError::CreateSignatureHeader)?,
        )
        .json(&duration)
        .send()
        .await
        .and_then(Response::error_for_status)
        .map_err(ApiError::from)?;

    let data =
        serde_json::from_slice::<ApiResponse<Data>>(&verify_response(handler, response).await?)
            .map_err(ApiError::from)?;

    Ok(data)
}

pub async fn exchange(
    handler: &Handler,
    signer: &Signer,
    meet_id: Uuid,
    messages: Vec<ExchangeMessage>,
) -> Result<()> {
    let client = client()?;

    let signature_header = signature_header(&messages, signer)?;

    let response = client
        .post(handler.add_path("v1/exchange"))
        .query(&[("meet_id", meet_id)])
        .header(
            HeaderName::from_static(SIGNATURE_HEADER_NAME),
            HeaderValue::from_str(&signature_header)
                .map_err(|_| ApiError::CreateSignatureHeader)?,
        )
        .json(&messages)
        .send()
        .await
        .and_then(Response::error_for_status)
        .map_err(ApiError::from)?;

    let _ = verify_response(handler, response).await?;

    Ok(())
}

fn signature_header<T: serde::Serialize>(value: &T, signer: &Signer) -> Result<String> {
    serde_json::to_vec(value)
        .map_err(|_| ApiError::CreateSignatureHeader.into())
        .and_then(|bytes| {
            let header = SignatureHeader {
                public_key: *signer.public_key(),
                account_id: signer.account().to_owned(),
                signature: signer.sign(&bytes),
            };

            serde_json::to_vec(&header).map_err(|_| ApiError::CreateSignatureHeader.into())
        })
        .map(|it| BASE64_STANDARD_NO_PAD.encode(it))
}

async fn verify_response(handler: &Handler, response: Response) -> Result<Vec<u8>> {
    let Some(header) = response.headers().get(HeaderName::from_static(SIGNATURE_HEADER_NAME)) else {
        return Err(ApiError::NoSignatureHeader.into())
    };

    let SignatureHeader { signature, .. } = header
        .to_str()
        .map_err(|_| ApiError::VerifySignatureHeader)
        .and_then(|header_str| {
            BASE64_STANDARD_NO_PAD
                .decode(header_str)
                .map_err(|_| ApiError::VerifySignatureHeader)
        })
        .and_then(|bytes| {
            serde_json::from_slice::<SignatureHeader>(&bytes)
                .map_err(|_| ApiError::VerifySignatureHeader)
        })?;

    let bytes = response
        .bytes()
        .await
        .map_err(|_| ApiError::VerifySignatureHeader)?;

    let server_public_key = view_server_key(handler).await?;
    server_public_key
        .verify(&bytes, &signature)
        .map_err(|_| ApiError::VerifySignatureHeader)?;

    Ok(bytes.to_vec())
}
